@@ -6,7 +6,7 @@ use utf8;
 
 package JSON::Relaxed::Parser;
 
-our $VERSION = "0.096";
+our $VERSION = "0.098";
 
 class JSON::Relaxed::Parser;
 
@@ -570,8 +570,7 @@ method is_comment_opener( $pretok ) {
     $pretok eq '//' || $pretok eq '/*';
 }
 
-sub min { $_[0] < $_[1] ? $_[0] : $_[1] }
-sub max { $_[0] > $_[1] ? $_[0] : $_[1] }
+use List::Util qw( min max uniqstr );
 
 method encode(%opts) {
     my $schema  = $opts{schema};
@@ -614,15 +613,16 @@ method encode(%opts) {
     }
 
     # Format a string value.
-    my $pr_string = sub ( $str ) {
+    my $pr_string = sub ( $str, $force = 0 ) {
 
 	# Reserved strings.
 	if ( !defined($str) ) {
 	    return "null";
 	}
 
-	if ( UNIVERSAL::isa( $str, 'JSON::Boolean' ) ) {
-	    return "".$str;	# force string result
+	if ( UNIVERSAL::isa( $str, 'JSON::Boolean' )
+	     || UNIVERSAL::isa( $str, 'JSON::PP::Boolean' ) ) {
+	    return (qw(false true))[$str];	# force string result
 	}
 
 	my $v = $str;
@@ -646,6 +646,8 @@ method encode(%opts) {
 	     || $v =~ $p_reserved
 	     || $v =~ $p_quotes
 	     || $v =~ /\s/
+	     || $v =~ /^(true|false|null)$/
+	     || !length($v)
 	   ) {
 	    if ( $v !~ /\"/ ) {
 		return '"' . $v . '"';
@@ -659,8 +661,8 @@ method encode(%opts) {
 	    return '"' . ($v =~ s/(["'`])/\\$1/rg) . '"';
 	}
 
-	# Just a string, potentially empty.
-	return length($v) ? $v : '""';
+	# Just a string.
+	return $v;
     };
 
     # Format an array value.
@@ -690,7 +692,7 @@ method encode(%opts) {
 	return $s;
     };
 
-    # Format a hask value.
+    # Format a hash value.
     my $pr_hash; $pr_hash = sub ( $rv, $level=0, $props = {} ) {
 	return "{}" unless keys(%$rv);
 
@@ -707,6 +709,9 @@ method encode(%opts) {
 	  ? @{ delete($rv->{" key order "}) }
 	  : sort(keys(%$rv));
 
+	# Dedup.
+	@ko = uniqstr(@ko);
+
 	my $ll = 0;
 	for ( @ko ) {
 	    # This may be wrong if \ escapes or combined keys are involved.
@@ -740,8 +745,12 @@ method encode(%opts) {
 	    $s .= $t;
 	    my $in = $comment ? "" : " " x max( 0, $ll-length($t) );
 
+	    # Handle object serialisation.
+	    my $r = UNIVERSAL::can( $v, "TO_JSON" ) // UNIVERSAL::can( $v, "FREEZE" );
+	    $r = $r ? $v->$r : $v;
+
 	    # Format the value.
-	    if ( ref($v) eq 'HASH' ) {
+	    if ( ref($r) eq 'HASH' ) {
 		# Make up and recurse.
 		if ( $pretty ) {
 		    $s .= $prpmode ? " " : " : ";
@@ -750,16 +759,16 @@ method encode(%opts) {
 		    $s .=  ":";
 		}
 
-		$s .= $pr_hash->( $v, $level+1, $props->{$k}->{properties} );
+		$s .= $pr_hash->( $r, $level+1, $props->{$k}->{properties} );
 	    }
 
-	    elsif ( ref($v) eq 'ARRAY' ) {
+	    elsif ( ref($r) eq 'ARRAY' ) {
 		$s .= $pretty ? "$in : " : ":";
-		$s .= $pr_array->( $v, $level+1, $props->{$k}->{items} );
+		$s .= $pr_array->( $r, $level+1, $props->{$k}->{items} );
 	    }
 
 	    elsif ( $pretty ) {
-		my $t = $pr_string->($v);
+		my $t = $pr_string->($r);
 		$s .= "$in : ";
 
 		# Break quoted strings that contain pseudo-newlines.
@@ -788,7 +797,7 @@ method encode(%opts) {
 		}
 	    }
 	    else {
-		$s .= ":" . $pr_string->($v) . ",";
+		$s .= ":" . $pr_string->($r) . ",";
 	    }
 	    $s .= "\n" if $pretty;
 	}
@@ -809,18 +818,23 @@ method encode(%opts) {
 	return $s;
     };
 
+    # Handle object serialisation.
+    my $r = UNIVERSAL::can( $rv, "TO_JSON" ) // UNIVERSAL::can( $rv, "FREEZE" );
+    $r = $r ? $rv->$r : $rv;
+
     # From here it is straight forward.
-    if ( ref($rv) eq 'HASH' ) {
-	$s .= $pr_hash->( $rv, $level, $props );
+    if ( ref($r) eq 'HASH' ) {
+	$s .= $pr_hash->( $r, $level, $props );
     }
-    elsif ( ref($rv) eq 'ARRAY' ) {
-	$s .= $pr_array->( $rv, $level );
+    elsif ( ref($r) eq 'ARRAY' ) {
+	$s .= $pr_array->( $r, $level );
     }
     else {
-	$s .= $pr_string->($rv);
+	$s .= $pr_string->($r);
     }
 
     # Final make-up.
+    $s =~ s/^ +$//gm;
     if ( $pretty && !$level ) {
 	$s =~ s/^\n*//s;
 	$s .= "\n" if $s !~ /\n$/;
@@ -1102,7 +1116,7 @@ method as_perl( %options ) {
 
     # null -> undef
     elsif ( $content eq "null" ) {
-	return;
+	return undef;
     }
 
     # Return as string.
@@ -1127,6 +1141,7 @@ package JSON::Boolean {
     sub _data_printer( $self, $ddp ) { "Bool($self)" }
 
     use overload '""'     => sub { ${$_[0]} ? "true" : "false" },
+		 "0+"     => sub { ${$_[0]} },
 		 "bool"   => sub { !!${$_[0]} },
 		 fallback => 1;
 
